-----------------------------------------------------------------------------
-- |
-- Copyright   :  (C) Dimitri Sabadie
-- License     :  BSD3
-- 
-- Maintainer  :  dimitri.sabadie@gmail.com
-- Stability   :  stable
-- Portability :  portable
--
-- This module re-exports anything about the 'MonadJournal' *typeclass*.
-----------------------------------------------------------------------------

module Control.Monad.Journal (
    module Control.Monad.Journal.Class
  ) where

import Control.Monad.Journal.Class
